/*
A GLUT program using OpenGL and Newton.
It uses the Newton physics API:
	http://www.newtondynamics.com/
	http://walaber.com/newton_wiki/index.php?title=Main_Page

It simulates some simple rocks.

Orion Sky Lawlor, olawlor@acm.org, 2009-04-09 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

#include "osl/random.h"
#include "osl/random_vector.h"
#include "osl/random.cpp"

/* Dr. Lawlor's OpenGL utilities */
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"

#include "ogl/minicam.h"

/* Newton Physics Engine: http://www.newtondynamics.com/ */
#define _NEWTON_USE_LIB 1 /* I want the statically linked library, no DLLs! */
#include "Newton.h"

NewtonBody *uiRock=NULL;

void standardPhysics(const NewtonBody *body)
{
	float mass,Ixx,Iyy,Izz;
	// Look up mass:
	NewtonBodyGetMassMatrix (body, &mass, &Ixx, &Iyy, &Izz);
	
	mat4 m(1.0);
	NewtonBodyGetMatrix(body,&m[0][0]);
	vec3 position(m[3]); /* translation column of the matrix */
	
	vec3 vel;
	NewtonBodyGetVelocity(body,vel);
	vec3 force=-mass * position; /* gravity points downward, force proportional to mass */
	
	if (body==uiRock) {
		float f=100.0; /* UI force magnitude */
		float targvel=1.0; /* velocity we try to reach via force */
		if (key_down[(int)'x']) force.x+=f*(targvel-vel.x);
		if (key_down[(int)'X']) force.x+=f*(-targvel-vel.x);
		if (key_down[(int)'y']) force.y+=f*(targvel-vel.y);
		if (key_down[(int)'Y']) force.y+=f*(-targvel-vel.y);
		if (key_down[(int)'z']) force.z+=f*(targvel-vel.z);
		if (key_down[(int)'Z']) force.z+=f*(-targvel-vel.z);
	}
	
	NewtonBodySetForce (body, force);
}





class Rock {
public:
	Rock(NewtonWorld *world_,osl::Random &r);
	void draw(void);
	NewtonWorld *world;
	NewtonBody *body;
	double sz; /* diameter of our object */
	bool isSphere;
	vec3 color;
	
	// add our forces and torques to rigid body
	static void PhysicsForceAndTorque (const NewtonBody* body);
};

Rock::Rock(NewtonWorld *world_,osl::Random &r)
	:world(world_)
{
	sz=0.01+0.2*r.nextFloat();
	NewtonCollision *c;
	if (r.nextInt(3)!=0) { /* make a box */
		c=NewtonCreateBox(world,sz,sz,sz,NULL);
		isSphere=false;
	} else { /* make a sphere */
		c=NewtonCreateSphere(world,sz,sz,sz,NULL);
		isSphere=true;
	}
	body=NewtonCreateBody(world,c);
	NewtonReleaseCollision(world,c);
	
	double rotNertia=0.01;
	NewtonBodySetMassMatrix(body,1,rotNertia,rotNertia,rotNertia);
	NewtonBodySetForceAndTorqueCallback(body,standardPhysics);
	
	mat4 m(1.0); // rock-to-world coordinates: orientation and position
	vec3 v(r.nextFloat(),r.nextFloat(),r.nextFloat()); // position
	v=(v-vec3(0.5f))*6;
	m[3][0]=v.x; /* spread objects apart */
	m[3][1]=v.y;
	m[3][2]=v.z;
	NewtonBodySetMatrix(body,&m[0][0]);
	
	color=vec3(r.nextFloat(),r.nextFloat(),r.nextFloat());
}
void Rock::draw(void)
{
	glPushMatrix();
	
	mat4 m(1.0);
	NewtonBodyGetMatrix(body,&m[0][0]);
	glMultMatrixf(&m[0][0]);
	
	glColor3fv(color);
	if (isSphere) glutSolidSphere(sz,6,5);
	else glutSolidCube(sz);
	
	if (body==uiRock) oglCameraAxes();
	
	glPopMatrix();
}

/* A set of interacting RobotPhysics and Rock objects */
class PhysicsWorld {
public:
	NewtonWorld *world;
	std::vector<Rock *> rocks;
	
	
	PhysicsWorld();
	void draw(void);
};

PhysicsWorld::PhysicsWorld()
{
	/* Fire up the physics */
	world = NewtonCreate(NULL,NULL);
	
	/* Add some robots and rocks */
	osl::Random48 r(1);
	for (int i=0;i<100;i++) {
		rocks.push_back(new Rock(world,r));
	}
	uiRock=rocks[0]->body;
}
void PhysicsWorld::draw(void)
{
	for (unsigned int i=0;i<rocks.size();i++)
		rocks[i]->draw();
}


PhysicsWorld *physicsworld=0; /*<- silly global variable */
GLUI *glui;  /* user interface */

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	/* Make sure there's a world out there */
	if (physicsworld==0) {
		physicsworld=new PhysicsWorld();
	}
	
	// Update physics, to keep up with real time
	static double sim_time=oglTimer(), dt=0.01;
	static double last_realtime=oglTimer();
	double cur_realtime=oglTimer();
	while (sim_time < cur_realtime) {
		sim_time+=dt;
		NewtonUpdate(physicsworld->world, dt);
		if (oglTimer()>last_realtime+0.1) break; /* just display */
	}
	last_realtime=cur_realtime;
	

	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);

	oglCameraLookAt(3.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	oglDirectionLight(GL_LIGHT0,oglVector3d(0.3,-0.2,1.0),
		oglVector3d(1,1,1),0.1,0.9,0.1);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	
	
	physicsworld->draw();
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
	// Repeatedly ask to be redrawn
	glutPostRedisplay();
}

int main(int argc, char** argv)
{
	camera=vec3(0.0,-3.0,0.0);
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("Newton Physics Demo: try XYZ keys");
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	oglCameraInit();
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
